/**********************************************************************************************/
/* FBL_BitSet.h																				  */
/*                                                                       					  */
/* Foundation Base Library ( FBL )		                                   					  */
/*                                                                       					  */
/* Copyright Paradigma, 1993-2002                                        					  */
/* All Rights Reserved.                                                 					  */
/**********************************************************************************************/

#ifndef __FBL_BitSet_h
   #define __FBL_BitSet_h

#if VSDK_FRAMEWORK_STYLE
#	include <VSDK/FBL_Set.h>
#else
#	include <FBL_Set.h>
#endif // VSDK_FRAMEWORK_STYLE

#if defined(__MWERKS__)
	#pragma options align=power
#endif


/**********************************************************************************************/
// Constants:

#define WORD_SIZE 		sizeof(int)
#define WORD_BITS_ON 	0xFFFFFFFF			// for sizeof(int) == 4


/**********************************************************************************************/
// Forward declarations:

class FBL_ArraySet;
class FBL_BitSet;
class FBL_BitSetIterator;
class FBL_SetIterator;


/**********************************************************************************************/
/*
	General purpose class.
	Implement Set of ulong numbers as BitSet.
	On creation you must specify top value.
*/

class FBL_BitSet : public FBL_Set
{
		friend class FBL_ArraySet;
		friend class FBL_BitSetIterator;

	public://///////////////////////////////////////////////////////////////////////////////////

		/* CONSTRUCTORS/DESTRUCTORS: */
		
FBL_EXP							FBL_BitSet( ID_TYPE	inMaxValue );						
FBL_EXP							FBL_BitSet( const FBL_BitSet& inSrc );
						
FBL_EXP							FBL_BitSet( 
										ID_TYPE			inMaxValue,	 
										const FBL_ArraySet*	inSet );
						
FBL_EXP virtual		    		~FBL_BitSet( void );


		/* Accessories: */
		
FBL_EXP	virtual FBL_SetIterator* GetIterator( void ) const;
		ID_TYPE					GetMaxValue( void ) const { return mMaxValue; }			


		/* SET OPERATIONS: */

								// Copy bytes from a given Set.
FBL_EXP	void					operator= ( const FBL_BitSet& inSet );

	
			/* Unary operators. Modify this set. */ 
	
FBL_EXP	void					Invert( void );	// invert all bits.
	
FBL_EXP	void					Union 		 		( const FBL_BitSet& inSet ); // OR
FBL_EXP	void					Intersection		( const FBL_BitSet& inSet ); // AND
FBL_EXP	void					Difference   		( const FBL_BitSet& inSet ); // AND NOT
FBL_EXP	void					SymmetricDifference	( const FBL_BitSet& inSet ); // XOR


			/* Binary operators. Create a new Set */
						
FBL_EXP friend	FBL_BitSet*		Union( 
									const FBL_BitSet& inSet1,
									const FBL_BitSet& inSet2 );
	
FBL_EXP friend	FBL_BitSet*		Intersection( 
									const FBL_BitSet& inSet1,
									const FBL_BitSet& inSet2 );
		
FBL_EXP friend FBL_BitSet*		Difference( 
									const FBL_BitSet& inSet1,
									const FBL_BitSet& inSet2 );
	

FBL_EXP friend FBL_BitSet*		SymmetricDifference( 
									const FBL_BitSet& inSet1,
									const FBL_BitSet& inSet2 );
	

		/* ITEM OPERATIONS:	 */

								// Returns true if inItem is in set.
		bool					Include( ID_TYPE inItem ) const;
	
								// Add/Remove value to/from set. Valid range 1..mMaxValue.
		void					Add 	( ID_TYPE inItem );
		void					Remove	( ID_TYPE inItem );
					
virtual ulong					GetCount( void ) const	{ return GetBits_1_Count(); }
FBL_EXP virtual bool			IsEmpty( void );
				
	
	public://///////////////////////////////////////////////////////////////////////////////////
 
FBL_EXP ID_TYPE 				FindNth_1_Bit( ulong inNth ) const;
FBL_EXP ID_TYPE					GetBits_1_Count( void ) const;
		ID_TYPE					GetBits_0_Count( void ) const { 
									return GetMaxValue() - GetBits_1_Count(); // FIXME. 2 iteration!
								}

								// Returns first and last nonzero bytes in set (range 0..N-1).
FBL_EXP void					GetNonZeroBounds( ID_TYPE& outFirst, ID_TYPE& outLast ) const;
		
		uchar*					GetArrayPtr( void ) const 		{ return mArray; 		}
		ulong					GetBytesInArray( void ) const	{ return mBytesInArray; }
																
								// Set all bits in set to 1/0.
		void					SetAll( void )		{ memset( mArray, 0xFF, mBytesInArray); }
		void					ClearAll( void ) 	{ memset( mArray, 0x00, mBytesInArray); }
								
								// Fill all bits in mArray starting from inStartValue by inItem.
FBL_EXP	void					SetBits  ( ID_TYPE inLeft = 1, ID_TYPE inRight = (ID_TYPE) -1 );
FBL_EXP	void					ClearBits( ID_TYPE inLeft = 1, ID_TYPE inRight = (ID_TYPE) -1 );

								// Fill all bytes in mArray with given BYTE value.
FBL_EXP	void					SetAllBytes( uchar inFillValue, ID_TYPE inStartByte = 0 );
								

	protected://////////////////////////////////////////////////////////////////////////////////

FBL_EXP	void 					InitSelf( ID_TYPE inMaxValue );
								
								// Calculate byte in mArray which contains bit of inItem.
		ulong					Loc( ID_TYPE inItem ) const { return inItem >> 3; }
								
								// Returns Mask. &7 - means cutting of low 3 bits from value.
		uchar					Mask( ID_TYPE inItem ) const { return uchar(1UL << (inItem & 0x07)); }
								

	protected://////////////////////////////////////////////////////////////////////////////////

		uchar*  				mArray;
		ulong					mBytesInArray;	// alligned to int.
								
		ID_TYPE					mMaxValue;
								
FBL_EXP	static 	uchar     		sMasks[8];
FBL_EXP	static	ulong 			sBitsCount[256];

};


/**********************************************************************************************/
inline bool FBL_BitSet::Include( ulong inItem ) const
{
	if( inItem && inItem <= mMaxValue ){
		--inItem;	// make inItem zero based.
		return bool( (mArray[ Loc(inItem) ] & Mask(inItem) ) != 0 );
	}	
	return false;
	
} /* FBL_BitSet::Include */


/**********************************************************************************************/
inline void FBL_BitSet::Add( ulong inItem )
{
	if( inItem && inItem <= mMaxValue ){
		inItem--;	// make inItem zero based.
		mArray[ Loc( inItem ) ] |= Mask( inItem );
	}
	
} /* FBL_BitSet::Add */


/**********************************************************************************************/
inline void FBL_BitSet::Remove( ulong inItem )
{
	if( inItem && inItem <= mMaxValue ){
		inItem--;
		mArray[ Loc( inItem ) ] &= (uchar) ~Mask( inItem );
	}
	
} /* FBL_BitSet::Remove */


/**********************************************************************************************/
inline void FBL_BitSet::SetAllBytes( uchar inFillValue, ID_TYPE inStartByte )
{
	uchar* p1  = mArray + inStartByte;
	memset( p1, inFillValue, mBytesInArray - inStartByte );				
	     
} /* FBL_BitSet::SetAllBytes */


/**********************************************************************************************/
#if defined(__MWERKS__)
	#pragma options align=reset
#endif

#endif
